#![allow(unused_variables)]

use std::collections::HashMap;

use crate::{
    checked::{self, Node, NodeKind},
    ir::{Block, Fn, Op},
    sym::Syms,
};

pub struct IrProgram {
    pub fns: Vec<Fn>,
    pub entry: u64,
    pub sym: Syms,
}

struct Compiler {
    syms: Syms,
    fns: Vec<Fn>,
}

impl Compiler {
    pub fn new(syms: Syms) -> Self {
        Self {
            syms,
            fns: Vec::new(),
        }
    }

    pub fn compile(mut self, ast: &[Node]) -> Result<IrProgram, ()> {
        let (blocks, local_count) = self.compile_fn_body(ast)?;
        let entry_fn = Fn {
            blocks,
            id: 0,
            arg_count: 0,
            local_count,
        };
        self.fns.push(entry_fn);
        Ok(IrProgram {
            fns: self.fns,
            sym: self.syms,
            entry: 0,
        })
    }

    fn compile_fn(&mut self, node: &Node) -> Result<Fn, ()> {
        let Node { kind: NodeKind::Fn { subject: _, params, return_typ: _, body, id }, .. } = node else {
            unreachable!()
        };
        let NodeKind::Block(body) = &body.kind  else { unreachable!() };
        let (blocks, local_count) = self.compile_fn_body(body)?;
        Ok(Fn {
            blocks,
            id: *id,
            arg_count: params.len() as i32,
            local_count,
        })
    }

    fn compile_fn_body(&mut self, body: &[Node]) -> Result<(Vec<Block>, i32), ()> {
        let mut blocks = Vec::new();
        let mut ops = Vec::<Op>::new();
        let mut local_count = 0;
        let mut local_map = HashMap::<u64, i32>::new();
        for node in body {
            let syms = self.syms.view(node.table_id);
            match &node.kind {
                NodeKind::Error => return Err(()),
                NodeKind::Id(id) => {
                    let sym = syms.get(*id).unwrap();
                    let local_id = local_map.get(&sym.uid).unwrap();
                    ops.push(Op::LoadLocal(*local_id));
                }
                NodeKind::Int(value) => match node.typ {
                    checked::Type::I32 => ops.push(Op::PushI32(*value as i32)),
                    checked::Type::U32 => ops.push(Op::PushU32(*value as u32)),
                    _ => unreachable!(),
                },
                NodeKind::String(_) => todo!(),
                NodeKind::Group(_) => todo!(),
                NodeKind::Block(_) => todo!(),
                NodeKind::Call { subject, args } => todo!(),
                NodeKind::If {
                    cond,
                    truthy,
                    falsy,
                } => todo!(),
                NodeKind::Loop { body } => todo!(),
                NodeKind::Break => todo!(),
                NodeKind::Assign { subject, value } => todo!(),
                NodeKind::Let { subject, value } => todo!(),
                NodeKind::Fn {
                    subject,
                    params,
                    return_typ,
                    body,
                    id,
                } => todo!(),
                NodeKind::Return { value } => todo!(),
                NodeKind::Param { subject, typ } => todo!(),
            }
        }
        blocks.push(Block { ops });
        Ok((blocks, local_count))
    }

    fn error<S: Into<String>>(&mut self, msg: S) {
        println!("ir compiler: {}", msg.into());
    }
}

#[test]
fn test_checker() {
    use crate::checker::{Checker, IdGen};
    use crate::parser::Parser;
    use pretty_assertions::assert_eq;
    use Op::*;

    struct SeqIdGen(u64);
    impl IdGen for SeqIdGen {
        fn new() -> Self {
            Self(0)
        }

        fn gen(&mut self) -> u64 {
            let v = self.0;
            self.0 += 1;
            v
        }
    }

    let compile = |text| {
        let mut checker = Checker::<SeqIdGen>::new_with_fn_id_gen();
        let checked = checker.check(&Parser::new(text).parse());
        let syms = checker.finish();
        let compiled = Compiler::new(syms).compile(&checked);
        compiled.map(|program| program.fns)
    };

    assert_eq!(
        compile("123;"),
        Ok(vec![Fn {
            blocks: vec![Block {
                ops: vec![PushI32(123)]
            }],
            id: 0,
            arg_count: 0,
            local_count: 0
        }])
    );
}
